/**
 * Copyright (c) 2010-2012 EBM WebSourcing, 2012-2018 Linagora
 * 
 * This program/library is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 2.1 of the License, or (at your
 * option) any later version.
 * 
 * This program/library is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License
 * for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program/library; If not, see http://www.gnu.org/licenses/
 * for the GNU Lesser General Public License version 2.1.
 */
package org.ow2.petals.launcher;

import java.io.File;
import java.io.FilenameFilter;
import java.net.MalformedURLException;
import java.net.URL;
import java.net.URLClassLoader;
import java.util.Arrays;

import org.ow2.petals.launcher.configuration.Configuration;
import org.ow2.petals.launcher.exception.PetalsLauncherException;

/**
 * A class loader that load jars *recursively* from the the lib directory and from the extension directory
 * 
 * @author Sebastien Andre - EBM WebSourcing
 */
public class PetalsContainerClassLoader extends URLClassLoader {

    // we have there: at least one number, followed by as much as we want numbers separated by a dot
    // and then optionally a dash and anything until there is the jar extension.
    private static final String PETALS_KERNEL_JAR_REGEXP = "petals-microkernel-(\\d+)(\\.(\\d+))*(-.*)?\\.jar";

    private static final FilenameFilter KERNEL_FILTER = new FilenameFilter() {
        @Override
        public boolean accept(final File dir, final String name) {
            return name.matches(PETALS_KERNEL_JAR_REGEXP);
        }
    };

    public PetalsContainerClassLoader(final Configuration config) throws PetalsLauncherException {
        super(new URL[] {}, PetalsContainerClassLoader.class.getClassLoader());

        try {

            // Add the kernel JAR to the classloader
            // -> it will also load the dependencies from its manifest
            // and make them visible in this current classloader
            final File[] kernelFiles = config.getLibDirectory().listFiles(KERNEL_FILTER);
            if (kernelFiles == null) {
                throw new PetalsLauncherException(config.getLibDirectory()
                        + " is not a valid directory: wo won't be able to find the microkernel jar.");
            } else if (kernelFiles.length < 1) {
                throw new PetalsLauncherException("No file matching the microkernel filename pattern (" + KERNEL_FILTER
                        + ") in the directory " + config.getLibDirectory());
            } else if (kernelFiles.length > 1) {
                throw new PetalsLauncherException("More than one file matching the microkernel filename pattern ("
                        + KERNEL_FILTER + ") in the directory " + config.getLibDirectory() + ": "
                        + Arrays.deepToString(kernelFiles));
            } else {
                this.addFile(kernelFiles[0]);
            }

            // Add the microkernel component implementations to the classloader
            // This is needed to load the registry implementation for example
            // TODO reference the dependencies from the manifest of the implementation?
            // NOTE: we must use recursion because there could be many subdirectories
            final File implDir = config.getImplementationsDirectory();
            if (implDir == null) {
                throw new PetalsLauncherException("Missing implementation directory in settings");
            } else if (!implDir.exists() || !implDir.isDirectory()) {
                throw new PetalsLauncherException("The provided implementation directory '" + implDir
                        + "' does not exists or is not a directory.");
            } else {
                this.addJarsInSubDirs(implDir);
            }

            // Add the microkernel extensions to the classloader
            // this is needed to load all the extensions
            // TODO have one directory per extension?
            // TODO reference the dependencies from the manifest of the extension?
            final File extDir = config.getExtensionsDirectory();
            if (extDir != null) {
                // it's ok if the directory does not exist
                if (extDir.exists()) {
                    if (extDir.isDirectory()) {
                        addJarsInSubDirs(extDir);
                    } else {
                        throw new PetalsLauncherException(
                                "The provided extension directory '" + extDir + "' is not a directory.");
                    }
                }
            }
        } catch (final PetalsLauncherException e) {
            throw e;
        } catch (final Exception e) {
            throw new PetalsLauncherException(e);
        }
    }

    private void addJarsInSubDirs(final File dir) throws MalformedURLException {
        for (final File subdir : dir.listFiles()) {
            if (subdir.isDirectory()) {
                for (File file : subdir.listFiles()) {
                    if (file.getName().endsWith(".jar")) {
                        addFile(file);
                    }
                }
            }
        }
    }

    private void addFile(final File f) throws MalformedURLException {
        super.addURL(f.toURI().toURL());
    }
}
