/**
 * Copyright (c) 2010-2012 EBM WebSourcing, 2012-2018 Linagora
 * 
 * This program/library is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 2.1 of the License, or (at your
 * option) any later version.
 * 
 * This program/library is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License
 * for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program/library; If not, see http://www.gnu.org/licenses/
 * for the GNU Lesser General Public License version 2.1.
 */
package org.ow2.petals.launcher;

import java.io.File;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.IOException;
import java.io.InputStream;
import java.net.MalformedURLException;
import java.net.URI;
import java.net.URISyntaxException;
import java.net.URL;
import java.util.List;
import java.util.Properties;
import java.util.logging.LogManager;
import java.util.logging.Logger;

import javax.xml.transform.TransformerFactory;

import org.apache.commons.cli.CommandLine;
import org.apache.commons.cli.CommandLineParser;
import org.apache.commons.cli.DefaultParser;
import org.apache.commons.cli.HelpFormatter;
import org.apache.commons.cli.Option;
import org.apache.commons.cli.Options;
import org.apache.commons.cli.ParseException;
import org.ow2.petals.basisapi.exception.PetalsException;
import org.ow2.petals.commons.log.Level;
import org.ow2.petals.jmx.api.api.JMXClient;
import org.ow2.petals.jmx.api.api.PetalsJmxApiFactory;
import org.ow2.petals.jmx.api.api.exception.DuplicatedServiceException;
import org.ow2.petals.jmx.api.api.exception.MissingServiceException;
import org.ow2.petals.launcher.api.server.PetalsServer;
import org.ow2.petals.launcher.api.server.PetalsServerFactory;
import org.ow2.petals.launcher.api.server.conf.ConfigurationProperties;
import org.ow2.petals.launcher.configuration.Configuration;
import org.ow2.petals.launcher.configuration.DefaultConfiguration;
import org.ow2.petals.launcher.exception.InvalidServerPropertiesFileException;
import org.ow2.petals.launcher.exception.NoContainerConfigurationAvailableException;
import org.ow2.petals.launcher.exception.PetalsLauncherException;
import org.ow2.petals.launcher.exception.ServerPropertiesFileNotFoundException;
import org.ow2.petals.launcher.exception.UncompleteServerConfigurationException;
import org.ow2.petals.topology.generated.Container;
import org.ow2.petals.topology.generated.Domain;
import org.ow2.petals.topology.generated.Topology;

import com.ebmwebsourcing.easycommons.lang.UncheckedException;
import com.ebmwebsourcing.easycommons.properties.PropertiesException;
import com.ebmwebsourcing.easycommons.properties.PropertiesHelper;

/**
 * Command Line Interface to start petals ESB server.
 * 
 * @author Sebastien Andre - EBM WebSourcing
 */
public class PetalsLauncher {

    public static final String LOGGER_NAME = "Petals.Launcher";

    /**
     * Footer part of the usage message
     */
    public static final String USAGE_FOOTER = "\nWhich evolution would you like on Petals? Share it! http://www.petalslink.com/feedback"
            + System.getProperty("line.separator");

    public static final String NAME = "Petals ESB Server";

    public static final String PETALS_HOME_PROPERTY_NAME = "petals.home";

    // font: smslant
    private static final String[] TITLE = new String[] {
            "     ___   ____ __   ___    __    ____   ____ ____ ___  ",
            "    / _ \\ / __// /_ / _ |  / /   / __/  / __// __// _ ) ",
            "   / ___// _/ / __// __ | / /__ _\\ \\   / _/ _\\ \\ / _  | ",
            "  /_/   /___/ \\__//_/ |_|/____//___/  /___//___//____/  ", "" };

    /**
     * The short name of the option setting the URL of the Petals container
     * configuration
     */
    protected static final String URL_SHORT_OPTION = "u";

    /**
     * The long name of the option setting the URL of the Petals container
     * configuration
     */
    private static final String URL_LONG_OPTION = "url";

    /**
     * The short name of the option printing help on command line
     */
    protected static final String HELP_SHORT_OPTION = "h";

    /**
     * The long name of the option printing help on command line
     */
    private static final String HELP_LONG_OPTION = "help";

    /**
     * The short name of the option printing the version of Petals
     */
    protected static final String VERSION_SHORT_OPTION = "V";

    /**
     * The long name of the option printing the version of Petals
     */
    private static final String VERSION_LONG_OPTION = "version";

    /**
     * The short name of the option printing the extended error messages
     * (stack-traces)
     */
    protected static final String ERROR_SHORT_OPTION = "e";

    /**
     * The long name of the option printing the extended error messages
     * (stack-traces)
     */
    private static final String ERROR_LONG_OPTION = "error";

    /**
     * The parameter to stop a remote container
     */
    protected static final String COMMAND_STOP = "stop";

    private final Configuration config;

    public PetalsLauncher(Configuration config) {
        this.config = config;
    }

    /**
     * Launch Petals ESB Server
     */
    protected void launchPetalsServer() throws PetalsLauncherException, IOException, PropertiesException {
        // ascii art :)
        for (final String line : TITLE) {
            System.out.println(line);
        }

        System.out.println("Initializing logging subsystem");
        final Logger log = this.initializeLoggingSubsystem();
        log.info("Logging system initialised");

        log.info("Initializing container's classloader...");
        final ClassLoader ccl = Thread.currentThread().getContextClassLoader();
        try (final PetalsContainerClassLoader petalsKernelClassLoader = new PetalsContainerClassLoader(config)) {
            Thread.currentThread().setContextClassLoader(petalsKernelClassLoader);

            final PetalsServer kernelInstance = PetalsServerFactory.getInstance(petalsKernelClassLoader);

            log.info("Initializing server...");
            kernelInstance.init(config.getServerLocalProperties(), config.getTopology(), config.getDataRootPath());

            log.info("Starting server...");
            kernelInstance.start();

            synchronized (kernelInstance) {
                while (kernelInstance.isRunning()) {
                    // the implementation will notify when it is stopped
                    kernelInstance.wait();
                }
            }
        } catch (Exception e) {
            log.log(Level.SEVERE, "Error while starting the server", e);
            if (e instanceof PetalsLauncherException) {
                throw (PetalsLauncherException) e;
            } else {
                // let's preserve the exception message so that is printed to the user
                throw new PetalsLauncherException(e.getMessage(), e);
            }
        } finally {
            Thread.currentThread().setContextClassLoader(ccl);
        }
    }

    /**
     * <p>
     * Initialize the logging system.
     * </p>
     * 
     * <p>
     * The configuration file of the JDK common logging system is retrieved from following ordered item:
     * <ol>
     * <li>the system property {@value ConfigurationProperties#LOGGING_SYSTEM_CONF_FILE_PROPERTY_NAME},</li>
     * <li>the configuration property {@value ConfigurationProperties#LOGGING_SYSTEM_CONF_FILE_PROPERTY_NAME} of the
     * container local configuration properties (ie. <code>server.properties</code>),</li>
     * <li>the file '<code>logger.properties</code>' contained in the directory containing the container local
     * configuration properties (ie. <code>server.properties</code>).</li>
     * <li>the file '<code>logger.properties</code>' in the system classpath.</li>
     * </ol>
     * </p>
     * 
     * <p>
     * Note: we use our own system property and not the default one of {@link LogManager} in order to get everything
     * initialized not before the moment this method is called. This is needed because we want the handler to be
     * instantiated while the context classloader is set to the kernel's, which means that if
     * PetalsPayloadDumperFileHandler is used, it will use a {@link TransformerFactory} instantiated from the kernel
     * classloader.
     * </p>
     */
    private Logger initializeLoggingSubsystem() throws IOException, PropertiesException {

        // Retrieve the configuration file of the JDK common logging system
        String configurationPath = System.getProperty(ConfigurationProperties.LOGGING_SYSTEM_CONF_FILE_PROPERTY_NAME);

        if (configurationPath == null) {
            // Try to find the configuration file set in the local container configuration properties
            configurationPath = config.getServerLocalProperties()
                    .getProperty(ConfigurationProperties.LOGGING_SYSTEM_CONF_FILE_PROPERTY_NAME);
        }

        if (configurationPath == null && config.getServerPropertiesUrl() != null) {
            configurationPath = new File(
                    new File(URI.create(config.getServerPropertiesUrl().toString())).getParentFile(),
                    ConfigurationProperties.DEFAULT_LOGGER_FILENAME).getAbsolutePath();
        }

        if (configurationPath == null) {
            System.out.println("Using default configuration file for log ("
                    + ConfigurationProperties.DEFAULT_LOGGER_FILENAME + " in the classpath)");
            // Retrieve the logging.properties path from the system classpath
            final URL configurationURL = ClassLoader.getSystemClassLoader()
                    .getResource("/" + ConfigurationProperties.DEFAULT_LOGGER_FILENAME);
            if (configurationURL == null) {
                throw new FileNotFoundException(
                        "Failed to reach the resource [" + ConfigurationProperties.DEFAULT_LOGGER_FILENAME + "]");
            }
            configurationPath = configurationURL.toString();
        }

        final File configurationFile = new File(configurationPath);

        // Load Petals customized log level
        Level.initialize();

        try (final InputStream fisInitial = new FileInputStream(configurationFile)) {
            final Properties existingProperties = new Properties(System.getProperties());
            existingProperties.putAll(config.getServerLocalProperties());
            try (final InputStream fisResolved = PropertiesHelper.resolvePropertiesForInputStream(fisInitial,
                    existingProperties)) {
                LogManager.getLogManager().readConfiguration(fisResolved);
                // let's request a logger to ensure handlers are loaded now and any problem will make an exception be
                // logged now and not at a future uncertain time.
                // Note: it won't stop the execution, just print on System.err!
                return Logger.getLogger(LOGGER_NAME);
            }
        }
    }

    protected void stopLocalPetalsServer() throws PetalsLauncherException {
        try {
            final Topology topology = config.getTopology();
            final String localContainerId = config.getLocalContainerId();
            final Domain domain = topology.getDomain();
            final List<Container> containers = domain.getContainer();
            for (final Container container : containers) {
                if (container.getName().equals(localContainerId)) {
                    final int rmiPort = (container.getJmxService() != null && container.getJmxService().getRmiPort() != 0)
                            ? container.getJmxService().getRmiPort() : 7700;
                    final String user = (container.getUser() != null) ? container.getUser() : "petals";
                    final String pwd = (container.getPassword() != null) ? container.getPassword() : "petals";
    
                    final JMXClient jmxClient = PetalsJmxApiFactory.getInstance().createJMXClient(container.getHost(),
                            rmiPort, user, pwd);
                    try {
                        jmxClient.getPetalsAdminServiceClient().stopContainer();
                    } finally {
                        jmxClient.disconnect();
                    }
                }
            }
        } catch (final UncompleteServerConfigurationException |
            DuplicatedServiceException | MissingServiceException | PetalsException e) {
            // let's preserve the exception message so that it is printed to the user
            throw new PetalsLauncherException(e.getMessage(), e);
        }
    }

    /**
     * Print versions informations
     */
    private static void printVersions(final CommandLine cmd) throws IOException, PetalsLauncherException,
            org.ow2.petals.launcher.api.server.exception.DuplicatedServiceException,
            org.ow2.petals.launcher.api.server.exception.MissingServiceException {
        try (final PetalsContainerClassLoader petalsKernelClassLoader = new PetalsContainerClassLoader(
                PetalsLauncher.getConfiguration(cmd))) {

            final PetalsServer kernelInstance = PetalsServerFactory.getInstance(petalsKernelClassLoader);
            final StringBuilder sb = new StringBuilder();
            sb.append(String.format("Petals ESB Container %s%n",
                    PetalsLauncher.getConfiguration(cmd).getPetalsESBContainerVersion()));
            sb.append(String.format("Petals ESB Launcher %s%n",
                    PetalsLauncher.class.getPackage().getImplementationVersion()));
            sb.append(String.format("%s %s%n", kernelInstance.getImplementationName(),
                    kernelInstance.getImplementationVersion()));
            sb.append(String.format("%s %s %s%n", System.getProperty("java.runtime.name"),
                    System.getProperty("java.runtime.version"), System.getProperty("java.vm.vendor")));
            sb.append(String.format("%s %s %s%n", System.getProperty("os.name"), System.getProperty("os.version"),
                    System.getProperty("os.arch")));

            System.out.println(sb.toString());
        }
    }

    /**
     * Print the usage message
     */
    private static void printUsageMessage() {
        final File jar;
        try {
            jar = new File(PetalsLauncher.class.getProtectionDomain().getCodeSource().getLocation()
                .toURI());
        } catch (final URISyntaxException e) {
            throw new UncheckedException(e);
        }

        final HelpFormatter formatter = new HelpFormatter();
        formatter.setWidth(USAGE_FOOTER.length());
        formatter.printHelp(NAME, "java -jar .../" + jar.getName()
                + " [--help] [--version] [--error] [--url <container-cfg-url>] [stop]",
                PetalsLauncher.createOptions(), "");

        System.out.println(" " + COMMAND_STOP + "           Stop the local Petals ESB container");
        System.out.println("");
        System.out.println(USAGE_FOOTER);
    }

    public static Options createOptions() {
        final Options options = new Options();

        options.addOption(Option.builder(HELP_SHORT_OPTION).longOpt(HELP_LONG_OPTION).hasArg(false).required(false)
                .desc("Print this help message and exit.").build());
        options.addOption(Option.builder(VERSION_SHORT_OPTION).longOpt(VERSION_LONG_OPTION).hasArg(false)
                .required(false).desc("Print the version number and exit.").build());
        options.addOption(Option.builder(ERROR_SHORT_OPTION).longOpt(ERROR_LONG_OPTION).hasArg(false).required(false)
                .desc("Print the extended error message (stack-trace).").build());
        options.addOption(
                Option.builder(URL_SHORT_OPTION).longOpt(URL_LONG_OPTION).hasArg(true).argName("container-cfg-url")
                        .required(false).desc("The URL of the Petals container configuration.").build());

        return options;
    }

    /**
     * The main entry point
     * 
     * @param args
     *            Arguments from the command line
     */
    public static void main(final String[] args) {
        System.exit(PetalsLauncher.run(args));
    }

    /**
     * Parse options and command line arguments
     * 
     * @param args
     *            Arguments from the command line
     */
    protected static int run(final String[] args) {
        boolean errorEnable = false;
        final CommandLineParser parser = new DefaultParser();

        try {
            final CommandLine cmd = parser.parse(PetalsLauncher.createOptions(), args);

            if (cmd.hasOption(ERROR_SHORT_OPTION)) {
                errorEnable = true;
            }

            if (cmd.hasOption(PetalsLauncher.HELP_SHORT_OPTION)) {
                PetalsLauncher.printUsageMessage();
            } else if (cmd.hasOption(PetalsLauncher.VERSION_SHORT_OPTION)) {
                PetalsLauncher.printVersions(cmd);
            } else {
                PetalsLauncher.startOrStopPetals(cmd);
            }
        } catch (final ParseException e) {
            System.err.println("ERROR: " + e.getMessage() + "\n");
            PetalsLauncher.printUsageMessage();
            return 1;
        } catch (final Exception e) {
            System.err.println("ERROR: " + e.getMessage() + "\n");
            if (errorEnable) {
                e.printStackTrace(System.err);
            }
            return 1;
        }

        return 0;
    }

    /**
     * Start or stop a Petals server according to the arguments on command line.
     * 
     * @param cmd
     *            The command line containing arguments
     */
    private static void startOrStopPetals(final CommandLine cmd)
            throws PetalsLauncherException, ParseException, IOException, PropertiesException {
        final String[] remainingArgs = cmd.getArgs();
        if (remainingArgs.length > 0) {
            if (remainingArgs[0].equalsIgnoreCase(COMMAND_STOP)) {
                new PetalsLauncher(PetalsLauncher.getConfiguration(cmd)).stopLocalPetalsServer();
            } else {
                throw new ParseException("Unknown option '" + remainingArgs[0]);
            }
        } else {
            new PetalsLauncher(PetalsLauncher.getConfiguration(cmd)).launchPetalsServer();
        }
    }

    private static Configuration getConfiguration(final CommandLine cmd) throws InvalidServerPropertiesFileException,
            UncompleteServerConfigurationException, NoContainerConfigurationAvailableException,
            ServerPropertiesFileNotFoundException {
        final URL serverLocalPropertiesUrl;
        if (cmd.hasOption(URL_SHORT_OPTION)) {
            final String serverLocalProperties = cmd.getOptionValue(URL_SHORT_OPTION);
            try {
                serverLocalPropertiesUrl = new URL(serverLocalProperties);
            } catch (final MalformedURLException e) {
                throw new InvalidServerPropertiesFileException(serverLocalProperties);
            }
        } else {
            serverLocalPropertiesUrl = null;
        }

        return new DefaultConfiguration(serverLocalPropertiesUrl);
    }
}
