/**
 * Copyright (c) 2013-2018 Linagora
 * 
 * This program/library is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 2.1 of the License, or (at your
 * option) any later version.
 * 
 * This program/library is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License
 * for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program/library; If not, see http://www.gnu.org/licenses/
 * for the GNU Lesser General Public License version 2.1.
 */
package org.ow2.petals.launcher;

import static org.junit.Assert.assertEquals;
import static org.junit.Assert.assertFalse;
import static org.junit.Assert.assertNotNull;
import static org.junit.Assert.assertTrue;

import java.io.IOException;
import java.net.URL;

import org.junit.Rule;
import org.junit.Test;
import org.junit.contrib.java.lang.system.ProvideSystemProperty;
import org.junit.contrib.java.lang.system.SystemErrRule;
import org.junit.contrib.java.lang.system.SystemOutRule;
import org.ow2.petals.jmx.api.api.PetalsJmxApiFactory;
import org.ow2.petals.jmx.api.mock.PetalsJmxApiFactoryMock;
import org.ow2.petals.launcher.api.server.PetalsServerFactory;
import org.ow2.petals.launcher.utest.PetalsServerUTest;

/**
 * Unit tests of {@link PetalsLauncher}
 * 
 * @author Christophe DENEUX - Linagora
 * 
 */
public class PetalsLauncherTest {

    @Rule
    public final ProvideSystemProperty props = new ProvideSystemProperty(PetalsJmxApiFactory.PROPERTY_NAME,
            PetalsJmxApiFactoryMock.class.getName()).and(PetalsServerFactory.PROPERTY_NAME,
                    PetalsServerUTest.class.getName());

    @Rule
    public final SystemOutRule systemOut = new SystemOutRule().muteForSuccessfulTests().enableLog();

    @Rule
    public final SystemErrRule systemErr = new SystemErrRule().muteForSuccessfulTests().enableLog();

    /**
     * <p>
     * Check the argument parsing against a command line containing:
     * </p>
     * <ul>
     * <li>the option to specify a configuration but without the url,</li>
     * <li>no help flag,</li>
     * <li>no version flag,</li>
     * <li>no error flag,</li>
     * <li>no extra parameters.</li>
     * </ul>
     * <p>
     * Expected results:
     * </p>
     * <ul>
     * <li>an error occurs about the missing URL.</li>
     * </ul>
     */
    @Test
    public final void configurationArgumentWithoutValue() {

        assertEquals(1, PetalsLauncher.run(new String[] { "-" + PetalsLauncher.URL_SHORT_OPTION }));

        final String error = systemErr.getLog();
        final String output = systemOut.getLog();
        assertFalse("No error message print", error.isEmpty());
        assertTrue("It's not the expected error message.",
                error.contains("Missing argument") && error.trim().endsWith(PetalsLauncher.URL_SHORT_OPTION));
        assertFalse("Stack trace is present in error message", error.contains(PetalsLauncher.class.getName()));
        assertFalse("Stack trace is present in error message", error.contains("Caused by"));
        assertTrue("Usage header is missing", output.contains(PetalsLauncher.NAME));
        assertTrue("Usage footer is missing", output.contains(PetalsLauncher.USAGE_FOOTER));

    }

    /**
     * <p>
     * Check the argument parsing against a command line containing:
     * </p>
     * <ul>
     * <li>the option to specify a configuration but with an unexisting url,</li>
     * <li>no help flag,</li>
     * <li>no version flag,</li>
     * <li>no error flag,</li>
     * <li>no extra parameters.</li>
     * </ul>
     * <p>
     * Expected results:
     * </p>
     * <ul>
     * <li>an error occurs about the missing URL.</li>
     * </ul>
     */
    @Test
    public final void startWithConfigurationArgumentWithUnexistingFileAndErrorFlagNotEnabled() {

        final String fileName = "unexisting-server.properties";
        assertEquals(1,
                PetalsLauncher.run(new String[] { "-" + PetalsLauncher.URL_SHORT_OPTION, "file:///" + fileName }));

        final String error = systemErr.getLog();
        final String output = systemOut.getLog();
        assertFalse("No error message print", error.isEmpty());
        assertTrue("It's not the expected error message.", error.contains(fileName));
        assertFalse("Stack trace is present in error message", error.contains(PetalsLauncher.class.getName()));
        assertFalse("Stack trace is present in error message", error.contains("Caused by"));
        assertFalse("Usage header is present", output.contains(PetalsLauncher.NAME));
        assertFalse("Usage footer is present", output.contains(PetalsLauncher.USAGE_FOOTER));

    }

    /**
     * <p>
     * Check the argument parsing against a command line containing:
     * </p>
     * <ul>
     * <li>the option to specify a configuration but with an unexisting url,</li>
     * <li>no help flag,</li>
     * <li>no version flag,</li>
     * <li>error flag enabled,</li>
     * <li>no extra parameters.</li>
     * </ul>
     * <p>
     * Expected results:
     * </p>
     * <ul>
     * <li>an error occurs about the missing URL.</li>
     * </ul>
     */
    @Test
    public final void startWithConfigurationArgumentWithUnexistingFileAndErrorFlagEnabled() {

        final String fileName = "unexisting-server.properties";
        assertEquals(1, PetalsLauncher.run(new String[] { "-" + PetalsLauncher.URL_SHORT_OPTION, "file:///" + fileName,
                "-" + PetalsLauncher.ERROR_SHORT_OPTION }));

        final String error = systemErr.getLog();
        final String output = systemOut.getLog();
        assertFalse("No error message print", error.isEmpty());
        assertTrue("It's not the expected error message.", error.contains(fileName));
        assertTrue("Stack trace is not present in error message", error.contains(PetalsLauncher.class.getName()));
        assertTrue("Stack trace is not present in error message", error.contains("Caused by"));
        assertFalse("Usage header is present", output.contains(PetalsLauncher.NAME));
        assertFalse("Usage footer is present", output.contains(PetalsLauncher.USAGE_FOOTER));

    }

    /**
     * <p>
     * Check the argument parsing against a command line containing:
     * </p>
     * <ul>
     * <li>the option to specify a configuration with an unexisting url,</li>
     * <li>no help flag,</li>
     * <li>no version flag,</li>
     * <li>error flag enabled,</li>
     * <li>no extra parameters,</li>
     * <li>with the stop command.</li>
     * </ul>
     * <p>
     * Expected results:
     * </p>
     * <ul>
     * <li>an error occurs about the missing URL.</li>
     * </ul>
     */
    @Test
    public final void stopWithConfigurationArgumentWithUnexistingFileAndErrorFlagEnabled() {

        final String fileName = "unexisting-server.properties";
        assertEquals(1, PetalsLauncher.run(new String[] { "-" + PetalsLauncher.URL_SHORT_OPTION, "file:///" + fileName,
                "-" + PetalsLauncher.ERROR_SHORT_OPTION, PetalsLauncher.COMMAND_STOP }));

        final String error = systemErr.getLog();
        final String output = systemOut.getLog();
        assertFalse("No error message print", error.isEmpty());
        assertTrue("It's not the expected error message.", error.contains(fileName));
        assertTrue("Stack trace is not present in error message", error.contains(PetalsLauncher.class.getName()));
        assertTrue("Stack trace is not present in error message", error.contains("Caused by"));
        assertFalse("Usage header is present", output.contains(PetalsLauncher.NAME));
        assertFalse("Usage footer is present", output.contains(PetalsLauncher.USAGE_FOOTER));

    }

    /**
     * <p>
     * Check the Petals launcher against a command line containing:
     * </p>
     * <ul>
     * <li>the option to specify a configuration with a valid url,</li>
     * <li>no help flag,</li>
     * <li>no version flag,</li>
     * <li>error flag enabled,</li>
     * <li>no extra parameters,</li>
     * <li>with the stop command.</li>
     * </ul>
     * <p>
     * Expected results: no error occurs
     * </p>
     */
    @Test
    public final void stopWithConfigurationArgumentWithValidFileAndErrorFlagEnabled() throws IOException {

        assertEquals(0,
                PetalsLauncher
                        .run(new String[] { "-" + PetalsLauncher.URL_SHORT_OPTION, getDefaultConfiguration().toString(),
                                "-" + PetalsLauncher.ERROR_SHORT_OPTION, PetalsLauncher.COMMAND_STOP }));

        final String error = systemErr.getLog();
        final String output = systemOut.getLog();
        assertTrue("Error message not empty", error.isEmpty());
        assertFalse("Usage header is present", output.contains(PetalsLauncher.NAME));
        assertFalse("Usage footer is present", output.contains(PetalsLauncher.USAGE_FOOTER));

    }

    /**
     * <p>
     * Check the Petals launcher against a command line containing:
     * </p>
     * <ul>
     * <li>no option to specify a configuration url,</li>
     * <li>no help flag,</li>
     * <li>no version flag,</li>
     * <li>error flag enabled,</li>
     * <li>no extra parameters,</li>
     * <li>with another unrecognized option.</li>
     * </ul>
     * <p>
     * Expected results: an error occurs about the unrecognized option
     * </p>
     */
    @Test
    public final void unrecognizedOption() throws IOException {

        final String unrecognizedOption = "-z";
        assertEquals(1,
                PetalsLauncher.run(new String[] { "-" + PetalsLauncher.ERROR_SHORT_OPTION, unrecognizedOption }));

        final String error = systemErr.getLog();
        final String output = systemOut.getLog();
        assertFalse("Error message not empty", error.isEmpty());
        assertTrue("Unrecognized option missing", error.contains(unrecognizedOption));
        assertTrue("Usage header is present", output.contains(PetalsLauncher.NAME));
        assertTrue("Usage footer is present", output.contains(PetalsLauncher.USAGE_FOOTER));

    }

    /**
     * <p>
     * Check the Petals launcher against a command line containing:
     * </p>
     * <ul>
     * <li>no option to specify a configuration url,</li>
     * <li>no help flag,</li>
     * <li>no version flag,</li>
     * <li>error flag enabled,</li>
     * <li>with unrecognized extra parameters.</li>
     * </ul>
     * <p>
     * Expected results: an error occurs about the unrecognized option
     * </p>
     */
    @Test
    public final void stopWithUnrecognizedOption() throws IOException {

        final String unrecognizedExtraParam = "unrecognized-extra-param";
        assertEquals(1,
                PetalsLauncher.run(new String[] { "-" + PetalsLauncher.ERROR_SHORT_OPTION, unrecognizedExtraParam }));

        final String error = systemErr.getLog();
        final String output = systemOut.getLog();
        assertFalse("Error message not empty", error.isEmpty());
        assertTrue("Unrecognized option missing", error.contains(unrecognizedExtraParam));
        assertTrue("Usage header is present", output.contains(PetalsLauncher.NAME));
        assertTrue("Usage footer is present", output.contains(PetalsLauncher.USAGE_FOOTER));

    }

    /**
     * <p>
     * Check the Petals launcher against a command line containing:
     * </p>
     * <ul>
     * <li>the option to print usage,</li>
     * <li>error flag enabled.</li>
     * </ul>
     * <p>
     * Expected results: no error occurs
     * </p>
     */
    @Test
    public final void printUsage() throws IOException {

        assertEquals(0, PetalsLauncher
                .run(new String[] { "-" + PetalsLauncher.HELP_SHORT_OPTION, "-" + PetalsLauncher.ERROR_SHORT_OPTION }));

        final String error = systemErr.getLog();
        final String output = systemOut.getLog();
        assertTrue("Error message not empty", error.isEmpty());
        assertTrue("Usage header isn't present", output.contains(PetalsLauncher.NAME));
        assertTrue("Usage footer isn't present", output.contains(PetalsLauncher.USAGE_FOOTER));

    }

    private static URL getDefaultConfiguration() {
        final URL defaultServerProperties = Thread.currentThread().getContextClassLoader()
                .getResource("default.server.properties");
        assertNotNull("Default configuration not found !!", defaultServerProperties);
        return defaultServerProperties;
    }

}
